/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.gemstone.gemfire.distributed.internal.membership.gms.auth;

import static com.gemstone.gemfire.distributed.ConfigurationProperties.*;
import static org.junit.Assert.*;
import static org.mockito.Mockito.*;

import java.security.Principal;
import java.util.Properties;

import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.contrib.java.lang.system.RestoreSystemProperties;
import org.junit.experimental.categories.Category;

import com.gemstone.gemfire.LogWriter;
import com.gemstone.gemfire.distributed.DistributedMember;
import com.gemstone.gemfire.distributed.internal.DistributionConfig;
import com.gemstone.gemfire.distributed.internal.membership.gms.ServiceConfig;
import com.gemstone.gemfire.distributed.internal.membership.gms.Services;
import com.gemstone.gemfire.internal.logging.InternalLogWriter;
import com.gemstone.gemfire.internal.security.GeodeSecurityUtil;
import com.gemstone.gemfire.security.AuthInitialize;
import com.gemstone.gemfire.security.AuthenticationFailedException;
import com.gemstone.gemfire.security.Authenticator;
import com.gemstone.gemfire.security.GemFireSecurityException;
import com.gemstone.gemfire.test.junit.categories.SecurityTest;
import com.gemstone.gemfire.test.junit.categories.UnitTest;

@Category({ UnitTest.class, SecurityTest.class })
public class GMSAuthenticatorJUnitTest {

  private String prefix;
  private Properties props;
  protected Properties securityProps;
  private Services services;
  private GMSAuthenticator authenticator;
  private DistributedMember member;

  @Rule
  public RestoreSystemProperties restoreSystemProperties = new RestoreSystemProperties();

  @Before
  public void setUp() throws Exception {
    prefix = getClass().getName() + "$";
    props = new Properties();
    securityProps = new Properties();
    authenticator = new GMSAuthenticator();

    services = mock(Services.class);
    InternalLogWriter securityLog = mock(InternalLogWriter.class);
    when(services.getSecurityLogWriter()).thenReturn(mock(InternalLogWriter.class));

    DistributionConfig distributionConfig = mock(DistributionConfig.class);
    when(distributionConfig.getSecurityProps()).thenReturn(securityProps);

    ServiceConfig serviceConfig = mock(ServiceConfig.class);
    when(serviceConfig.getDistributionConfig()).thenReturn(distributionConfig);

    services = mock(Services.class);
    when(services.getSecurityLogWriter()).thenReturn(securityLog);
    when(services.getConfig()).thenReturn(serviceConfig);

    authenticator.init(services);
    GeodeSecurityUtil.initSecurity(securityProps);
    member = mock(DistributedMember.class);
  }

  @Test
  public void testGetCredentialNormal() throws Exception {
    props.setProperty(SECURITY_PEER_AUTH_INIT, prefix + "TestAuthInit2.create");
    TestAuthInit2 auth = new TestAuthInit2();
    assertFalse(auth.isClosed());
    TestAuthInit2.setAuthInitialize(auth);
    Properties credential = authenticator.getCredentials(member, props);
    assertTrue(props == credential);
    assertTrue(auth.isClosed());
    assertTrue(TestAuthInit2.getCreateCount() == 1);
  }

  @Test
  public void testGetCredentialWithNoAuth() throws Exception {
    Properties credential = authenticator.getCredentials(member, props);
    assertNull(credential);
  }

  @Test
  public void testGetCredentialWithEmptyAuth() throws Exception {
    props.setProperty(SECURITY_PEER_AUTH_INIT, "");
    Properties credential = authenticator.getCredentials(member, props);
    assertNull(credential);
  }

  @Test
  public void testGetCredentialWithNotExistAuth() throws Exception {
    props.setProperty(SECURITY_PEER_AUTH_INIT, prefix + "NotExistAuth.create");
    verifyNegativeGetCredential(props, "Instance could not be obtained");
  }

  @Test
  public void testGetCredentialWithNullAuth() throws Exception {
    props.setProperty(SECURITY_PEER_AUTH_INIT, prefix + "TestAuthInit1.create");
    verifyNegativeGetCredential(props, "Instance could not be obtained");
  }

  @Test
  public void testGetCredentialWithInitError() throws Exception {
    props.setProperty(SECURITY_PEER_AUTH_INIT, prefix + "TestAuthInit3.create");
    verifyNegativeGetCredential(props, "expected init error");
  }

  @Test
  public void testGetCredentialWithError() throws Exception {
    props.setProperty(SECURITY_PEER_AUTH_INIT, prefix + "TestAuthInit4.create");
    verifyNegativeGetCredential(props, "expected get credential error");
  }

  private void verifyNegativeGetCredential(Properties props, String expectedError) throws Exception {
    try {
      authenticator.getCredentials(member, props);
      fail("should catch: " + expectedError);
    } catch (GemFireSecurityException expected) {
      assertTrue(expected.getMessage().startsWith(expectedError));
    }
  }

  @Test
  public void testAuthenticatorNormal() throws Exception {
    props.setProperty(SECURITY_PEER_AUTHENTICATOR, prefix + "TestAuthenticator4.create");
    GeodeSecurityUtil.initSecurity(props);
    TestAuthenticator4 auth = new TestAuthenticator4();
    assertFalse(auth.isClosed());
    TestAuthenticator4.setAuthenticator(auth);
    String result = authenticator.authenticate(member, props, props, member);
    assertNull(result);
    assertTrue(auth.isClosed());
    assertTrue(TestAuthenticator4.getCreateCount() == 1);
  }

  @Test
  public void testAuthenticatorWithNoAuth() throws Exception {
    String result = authenticator.authenticate(member, props, props, member);
    assertNull(result);
  }

  @Test
  public void testAuthenticatorWithEmptyAuth() throws Exception {
    props.setProperty(SECURITY_PEER_AUTHENTICATOR, "");
    GeodeSecurityUtil.initSecurity(props);
    String result = authenticator.authenticate(member, props, props, member);
    assertNull(result);
  }

  @Test
  public void testAuthenticatorWithNotExistAuth() throws Exception {
    props.setProperty(SECURITY_PEER_AUTHENTICATOR, prefix + "NotExistAuth.create");
    GeodeSecurityUtil.initSecurity(props);
    verifyNegativeAuthenticate(props, props, "Authentication failed. See coordinator");
  }

  @Test
  public void testAuthenticatorWithNullAuth() throws Exception {
    props.setProperty(SECURITY_PEER_AUTHENTICATOR, prefix + "TestAuthenticator1.create");
    GeodeSecurityUtil.initSecurity(props);
    verifyNegativeAuthenticate(props, props, "Authentication failed. See coordinator");
  }

  @Test
  public void testAuthenticatorWithNullCredential() throws Exception {
    props.setProperty(SECURITY_PEER_AUTHENTICATOR, prefix + "TestAuthenticator1.create");
    GeodeSecurityUtil.initSecurity(props);
    verifyNegativeAuthenticate(null, props, "Failed to find credentials from");
  }

  @Test
  public void testAuthenticatorWithAuthInitFailure() throws Exception {
    props.setProperty(SECURITY_PEER_AUTHENTICATOR, prefix + "TestAuthenticator2.create");
    GeodeSecurityUtil.initSecurity(props);
    verifyNegativeAuthenticate(props, props, "Authentication failed. See coordinator");
  }

  @Test
  public void testAuthenticatorWithAuthFailure() throws Exception {
    props.setProperty(SECURITY_PEER_AUTHENTICATOR, prefix + "TestAuthenticator3.create");
    GeodeSecurityUtil.initSecurity(props);
    verifyNegativeAuthenticate(props, props, "Authentication failed. See coordinator");
  }

  void verifyNegativeAuthenticate(Properties credential, Properties props, String expectedError) throws Exception {
    String result = authenticator.authenticate(member, credential, props, member);
    assertTrue(result, result.startsWith(expectedError));
  }

  public static class TestAuthInit1 implements AuthInitialize {
    public static AuthInitialize create() {
      return null;
    }
    @Override
    public void init(LogWriter systemLogger, LogWriter securityLogger) throws AuthenticationFailedException {
    }
    @Override
    public Properties getCredentials(Properties props, DistributedMember server, boolean isPeer) throws AuthenticationFailedException {
      throw new AuthenticationFailedException("expected get credential error");
    }
    @Override
    public void close() {
    }
  }

  public static class TestAuthInit2 extends TestAuthInit1 {

    private static TestAuthInit2 instance = null;
    private static int createCount = 0;

    boolean closed = false;

    public static void setAuthInitialize(TestAuthInit2 auth) {
      instance = auth;
    }
    public static AuthInitialize create() {
      createCount ++;
      return instance;
    }
    @Override
    public Properties getCredentials(Properties props, DistributedMember server, boolean isPeer) throws AuthenticationFailedException {
      return props;
    }
    @Override
    public void close() {
      closed = true;
    }
    public boolean isClosed() {
      return closed;
    }
    public static int getCreateCount() {
      return createCount;
    }
  }

  // used by reflection by test
  public static class TestAuthInit3 extends TestAuthInit1 {
    public static AuthInitialize create() {
      return new TestAuthInit3();
    }
    @Override
    public void init(LogWriter systemLogger, LogWriter securityLogger) throws AuthenticationFailedException {
      throw new AuthenticationFailedException("expected init error");
    }
  }

  public static class TestAuthInit4 extends TestAuthInit1 {
    public static AuthInitialize create() {
      return new TestAuthInit4();
    }
  }

  public static class TestAuthenticator1 implements Authenticator {
    public static Authenticator create() {
      return null;
    }
    @Override
    public void init(Properties securityProps, LogWriter systemLogger, LogWriter securityLogger) throws AuthenticationFailedException {
    }
    @Override
    public Principal authenticate(Properties props, DistributedMember member) throws AuthenticationFailedException {
      return null;
    }
    @Override
    public void close() {
    }
  }

  public static class TestAuthenticator2 extends TestAuthenticator1 {
    public static Authenticator create() {
      return new TestAuthenticator2();
    }
    @Override
    public void init(Properties securityProps, LogWriter systemLogger, LogWriter securityLogger) throws AuthenticationFailedException {
      throw new AuthenticationFailedException("expected init error");
    }
  }

  public static class TestAuthenticator3 extends TestAuthenticator1 {
    public static Authenticator create() {
      return new TestAuthenticator3();
    }
    @Override
    public Principal authenticate(Properties props, DistributedMember member) throws AuthenticationFailedException {
      throw new AuthenticationFailedException("expected authenticate error");
    }
  }

  public static class TestAuthenticator4 extends TestAuthenticator1 {

    private static Authenticator instance = null;
    private static int createCount = 0;

    private boolean closed = false;

    public static void setAuthenticator(Authenticator auth) {
      instance = auth;
    }
    public static Authenticator create() {
      createCount ++;
      return instance;
    }
    @Override
    public Principal authenticate(Properties props, DistributedMember member) throws AuthenticationFailedException {
      return null;
    }
    @Override
    public void close() {
      closed = true;
    }
    public boolean isClosed() {
      return closed;
    }
    public static int getCreateCount() {
      return createCount;
    }
  }
}
